"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const prompt_for_input_1 = require("../../util/prompt-for-input");
const fs = require('fs');
const path = require('path');
const child_process_1 = require("child_process");
const environment_1 = require("../../environment");
const AdmZip = require('adm-zip');
const du = require('du');
const winston = require('winston');
const parseStringSync = require('xml2js-parser').parseStringSync;
const rimraf = require('rimraf');
const alwayscli_1 = require("@alwaysai/alwayscli");
const validate_dataset_1 = require("./validate-dataset");
const constants_1 = require("./constants");
const resize_1 = require("./resize");
const tar = require("tar");
const model_package_cache_1 = require("../../util/model-package-cache");
const path_1 = require("path");
const child_spawner_1 = require("../../util/spawner/child-spawner");
const prune_1 = require("../model/prune");
const yes_cli_input_1 = require("../../cli-inputs/yes-cli-input");
// Runs command with shell=true, necessary for commands that require spaces
async function runForegroundSync(cmd) {
    winston.debug(cmd);
    child_process_1.spawnSync(cmd, [], {
        stdio: 'inherit',
        shell: true,
    });
}
function createTrainJson(steps, classes, labels, batchSize, id, fromCheckPoint) {
    const output = {
        numSteps: steps,
        numClasses: classes,
        labels,
        batchSize,
        id,
        fromCheckPoint,
    };
    fs.writeFileSync('training-temp/alwaysai.train.json', JSON.stringify(output, null, 4));
}
function validateLabels(annotationsDir, labels) {
    const annotationFilesNames = fs.readdirSync(annotationsDir);
    // Should create annotationSet before this for loop and check labels after
    const annotationLabels = new Set();
    annotationFilesNames.forEach((annotationFileName) => {
        // Increment FileName Field in Annotation File
        const annotationJson_ = parseStringSync(fs.readFileSync(path.join(annotationsDir, annotationFileName)));
        const objectList = annotationJson_.annotation.object;
        objectList.forEach((object) => {
            annotationLabels.add(object.name[0]);
        });
    });
    winston.info(`Available labels for dataset:\n  ${Array.from(annotationLabels).join('\n  ')}`);
    labels.forEach(label => {
        if (!annotationLabels.has(label)) {
            throw `Label "${label}" not found in any annotation files Are you sure you typed them in correctly? Annotation file labels: ${Array.from(annotationLabels)}`;
        }
    });
}
function generateBatchSize() {
    let batchSize = 1;
    const numImages = fs.readdirSync(`training-temp/${constants_1.ImageDirName}`).length;
    let bytesNeeded = constants_1.PixelWidth * constants_1.PixelHeight * constants_1.BytesPerPixel;
    while (true) {
        const heap = process.memoryUsage();
        const availBytes = heap.heapTotal - heap.heapUsed;
        if (batchSize * 2 < numImages / 2 && bytesNeeded * 2 < availBytes) {
            bytesNeeded *= 2;
            batchSize *= 2;
        }
        else
            break;
    }
    return batchSize;
}
function getModelVersion(modelPath) {
    let highestVersion = 0.0;
    fs.readdirSync(modelPath).forEach(model => {
        const version = parseFloat(model.split('.tar')[0]);
        if (version > highestVersion) {
            highestVersion = version;
        }
    });
    return highestVersion;
}
async function checkCacheSize(modelPath, id, ask) {
    const sizeInBytes = await du(modelPath);
    const sizeInMB = sizeInBytes / 1000000;
    // Ask to prune after 500MB
    if (sizeInMB >= 500) {
        if (ask) {
            const message = `Your model cache is over ${sizeInMB} MB, Would you like to clear it?\n You can always clear manually using aai model prune`;
            const { confirmed } = await prompt_for_input_1.promptForInput({
                purpose: `to confirm if you want to clear cache`,
                questions: [
                    {
                        type: 'confirm',
                        name: 'confirmed',
                        message,
                        initial: true,
                    },
                ],
            });
            if (confirmed) {
                prune_1.modelPrune.action(id, { ignore: undefined }, undefined);
            }
        }
        else {
            prune_1.modelPrune.action(id, { ignore: undefined }, undefined);
        }
    }
}
const datasetTrainFunc = () => alwayscli_1.CliLeaf({
    name: 'train',
    description: 'Produce object detection model for dataset',
    hidden: !environment_1.ALWAYSAI_SHOW_HIDDEN,
    positionalInput: alwayscli_1.CliStringArrayInput({
        placeholder: '</path/to/dataset1.zip> </path/to/dataset2.zip> ...',
        required: true,
    }),
    namedInputs: {
        labels: alwayscli_1.CliStringArrayInput({
            description: 'Dataset labels to train on',
            required: true,
        }),
        numSteps: alwayscli_1.CliNumberInput({
            description: 'Train for x number of steps',
            required: true,
        }),
        id: alwayscli_1.CliStringInput({
            description: 'ID of the model produced ex: alwaysai/dog_detector',
            required: true,
        }),
        hardware: alwayscli_1.CliStringInput({
            description: 'CPU or GPU. Default is set to CPU.',
            required: false,
            defaultValue: 'CPU',
        }),
        yes: yes_cli_input_1.yesCliInput,
        'continue-from-version': alwayscli_1.CliStringInput({
            description: 'Cached version of model to ex: 0.1',
            required: false,
        }),
    },
    async action([...args], opts) {
        winston.info(`Datasets\n  ${args.join('\n  ')}`);
        winston.info(`Labels\n  ${opts['labels'].join('\n  ')}`);
        if (fs.existsSync('training-temp')) {
            rimraf.sync('training-temp');
        }
        if (args.length === 1) {
            await resize_1.datasetResize.action(undefined, {
                'target-dir': args[0],
                'output-dir': undefined,
                width: constants_1.PixelWidth,
                height: constants_1.PixelHeight,
            }, undefined);
            const zip = new AdmZip(`resized_${path.basename(args[0])}`);
            zip.extractAllTo('training-temp', true);
            rimraf.sync(`resized_${path.basename(args[0])}`);
            validate_dataset_1.validateDataset('training-temp');
        }
        else {
            fs.mkdirSync('training-temp');
            await runForegroundSync(`aai dataset merge ${args.join(' ')}`);
            await resize_1.datasetResize.action(undefined, {
                'target-dir': 'merged-dataset.zip',
                'output-dir': undefined,
                width: constants_1.PixelWidth,
                height: constants_1.PixelHeight,
            }, undefined);
            const zip = new AdmZip('resized_merged-dataset.zip');
            zip.extractAllTo('training-temp', true);
            rimraf.sync('resized_merged-dataset.zip');
        }
        const labels = opts['labels'];
        validateLabels(`training-temp/${constants_1.AnnotationDirName}`, labels);
        const batchSize = generateBatchSize();
        winston.info(`Batch Size : ${batchSize}\n`);
        const id = opts['id'];
        let continueFromCheckpoint = false;
        if (opts['continue-from-version']) {
            const checkpointModelPackagePath = path.join(model_package_cache_1.ModelPackageDirectory(), id, `${opts['continue-from-version']}.tar.gz`);
            if (!fs.existsSync(checkpointModelPackagePath)) {
                throw new alwayscli_1.CliTerseError(`No model package found ${id} version ${opts['continue-from-version']}`);
            }
            const tempCheckPointPath = path.join('training-temp', 'checkpointFiles');
            fs.mkdirSync(tempCheckPointPath);
            await tar.extract({
                file: checkpointModelPackagePath,
                cwd: tempCheckPointPath,
                strict: true,
            });
            const files_ = fs.readdirSync(path.join(tempCheckPointPath, path.basename(id)));
            const ckptFiles = [];
            files_.forEach((f) => {
                if (f.startsWith('model.ckpt-')) {
                    ckptFiles.push(f);
                }
            });
            if (ckptFiles.length === 0) {
                throw new alwayscli_1.CliTerseError(`The model ${id} version ${opts['continue-from-version']} doesn't have any checkpoint files so we cannot continue from it.`);
            }
            continueFromCheckpoint = true;
        }
        createTrainJson(opts['numSteps'], labels.length, labels, batchSize, id, continueFromCheckpoint);
        const hardware = opts['hardware'];
        const cpuCount = Number(await child_spawner_1.ChildSpawner({ path: '.' }).run({
            exe: 'sh',
            args: ['-c', 'docker info | grep CPUs | grep -oE "[1-9][0-9]{0,2}" | head -n1'],
        }));
        if (hardware === 'GPU') {
            await runForegroundSync(`docker run -it --cpus ${cpuCount -
                1} --gpus all -v $\{PWD}/training-temp:/alwaysai alwaysai/train:gpu`);
        }
        else {
            await runForegroundSync(`docker run -it --cpus ${cpuCount -
                1} -v $\{PWD}/training-temp:/alwaysai alwaysai/train`);
        }
        if (!fs.existsSync(`training-temp/${id}`)) {
            throw new alwayscli_1.CliTerseError(`Exported model not found training-temp/${id}`);
        }
        const cacheDest = path.join(model_package_cache_1.ModelPackageDirectory(), id);
        if (!fs.existsSync(cacheDest)) {
            fs.mkdirSync(cacheDest);
        }
        else {
            await checkCacheSize(cacheDest, id, !opts['yes']);
        }
        const exportedModelDir = `training-temp/${id}`;
        const cachedModelVersion = getModelVersion(cacheDest);
        const version = cachedModelVersion + 0.1;
        const packageFile = path.join(cacheDest, `${version.toFixed(1)}.tar.gz`);
        fs.writeFileSync(packageFile, '');
        // Note: The TypeScript types on tar.create are not 100% accurate
        await tar.create({ cwd: path_1.dirname(exportedModelDir), gzip: true, file: packageFile }, [path_1.basename(exportedModelDir)]);
        winston.info(`Created model ${id} version ${version.toFixed(1)}. Add to your app using aai app models add ${id} or if already added to your app use aai app models update to use to this new model`);
    },
});
exports.datasetTrain = datasetTrainFunc();
exports.datasetTrainAlias = datasetTrainFunc();
exports.datasetTrainAlias.description = 'Alias for dataset train';
//# sourceMappingURL=train.js.map