"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const fs_1 = require("fs");
const node_fetch_1 = require("node-fetch");
const cloud_api_1 = require("@alwaysai/cloud-api");
const coded_error_1 = require("@carnesen/coded-error");
const tar = require("tar");
const path = require('path');
const fs = require('fs');
const rimraf = require("rimraf");
const rpc_client_1 = require("./rpc-client");
const authentication_client_1 = require("./authentication-client");
const model_json_file_1 = require("./model-json-file");
const util_1 = require("util");
const tempy = require("tempy");
const system_id_1 = require("./system-id");
const model_package_cache_1 = require("./model-package-cache");
const winston = require("winston");
const download_model_package_to_cache_1 = require("./download-model-package-to-cache");
function compareModelVersions(a, b) {
    const aVersion = parseFloat(a.split('.tar')[0]);
    const bVersion = parseFloat(b.split('.tar')[0]);
    if (aVersion < bVersion) {
        return -1;
    }
    if (aVersion > bVersion) {
        return 1;
    }
    return 0;
}
function getLatestModel(id) {
    const modelDirectory = path.join(model_package_cache_1.ModelPackageDirectory(), id);
    let modelPathList;
    try {
        modelPathList = fs.readdirSync(modelDirectory);
    }
    catch (error) {
        throw new Error(`Model ${id} not found in cache`);
    }
    modelPathList.sort(compareModelVersions);
    const latestModel = modelPathList.pop();
    return path.join(model_package_cache_1.ModelPackageDirectory(), id, latestModel);
}
function getVersionFromModelPath(path_) {
    return parseFloat(path.basename(path_).split('.tar')[0]);
}
async function extractConfigFromCachedModel(id, destDir) {
    const lateModelPath = getLatestModel(id);
    await tar.extract({ file: lateModelPath, cwd: destDir });
}
exports.modelPackageCloudClient = {
    async download(id, version) {
        const { uuid } = await rpc_client_1.CliRpcClient().getModelVersion({ id, version });
        const modelPackageUrl = ModelPackageUrl(uuid);
        const authorizationHeader = await authentication_client_1.CliAuthenticationClient().getAuthorizationHeader();
        const response = await node_fetch_1.default(modelPackageUrl, {
            method: 'GET',
            headers: Object.assign(Object.assign({}, authorizationHeader), { 'Content-Length': '0' }),
        });
        await throwIfNotOk(response);
        return response.body;
    },
    async publish(id) {
        let modelPackagePath;
        let modelJson;
        let originalVersion;
        let id_ = id;
        if (!id) {
            // Get the model config from current dir
            // tar the package
            const modelConfigDir = process.cwd();
            modelJson = model_json_file_1.ModelJsonFile(modelConfigDir);
            modelPackagePath = tempy.file();
            await tar.create({ cwd: path_1.dirname(modelConfigDir), gzip: true, file: modelPackagePath }, [path_1.basename(process.cwd())]);
            id_ = modelJson.id;
        }
        else {
            // Extract model config to temp dir and read
            // Define latest model for id as model package to publish
            const tempDir = tempy.directory();
            const modelConfigDir = path.join(tempDir, id.split('/').pop());
            await extractConfigFromCachedModel(id, tempDir);
            modelJson = model_json_file_1.ModelJsonFile(modelConfigDir).read();
            rimraf.sync(tempDir);
            modelPackagePath = getLatestModel(id);
            originalVersion = getVersionFromModelPath(modelPackagePath);
            id_ = id;
        }
        const rpcClient = rpc_client_1.CliRpcClient();
        const { uuid } = await rpcClient.createModelVersion(modelJson);
        const authorizationHeader = await authentication_client_1.CliAuthenticationClient().getAuthorizationHeader();
        const stats = await util_1.promisify(fs_1.stat)(modelPackagePath);
        const modelPackageStream = fs_1.createReadStream(modelPackagePath);
        const response = await node_fetch_1.default(ModelPackageUrl(uuid), {
            method: 'PUT',
            headers: Object.assign({ 'Content-Type': 'application/gzip', 'Content-Length': stats.size.toString() }, authorizationHeader),
            body: modelPackageStream,
        });
        await throwIfNotOk(response);
        await rpcClient.finalizeModelVersion(uuid);
        const { version: latestVersion } = await rpc_client_1.CliRpcClient().getModelVersion({ id: id_ });
        if (!id) {
            winston.info(`Published ${id_} version ${latestVersion}`);
        }
        else {
            winston.info(`Publishing from cache ${id_} version ${originalVersion}, Published to model catalog as ${id_} version ${latestVersion}`);
        }
        await download_model_package_to_cache_1.downloadModelPackageToCache(id_, latestVersion);
        return uuid;
    },
};
function ModelPackageUrl(uuid) {
    const cloudApiUrl = cloud_api_1.CloudApiUrl(system_id_1.getSystemId());
    return `${cloudApiUrl}${cloud_api_1.CLOUD_API_MODEL_VERSION_PACKAGES_PATH}/${uuid}`;
}
async function throwIfNotOk(response) {
    if (!response.ok) {
        const parsed = await response.json();
        const { message = `Server responded ${response.status} ("${response.statusText}")`, code, data, } = parsed;
        throw new coded_error_1.CodedError(message, code, data);
    }
}
//# sourceMappingURL=model-package-cloud-client.js.map