import {
  CliLeaf,
  CliStringInput,
  CliTerseError,
  CliNumberInput,
} from '@alwaysai/alwayscli';
import { AnnotationDirName, ImageDirName, PixelHeight, PixelWidth } from './constants';

const sharp = require('sharp');
const fs = require('fs');
const xml2js = require('xml2js');
const winston = require('winston');
const admzip = require('adm-zip');
const rimraf = require('rimraf');
const path = require('path');

const parseBoundingBox = function(boundingBox: any) {
  return {
    xmin: boundingBox.bndbox[0].xmin[0],
    xmax: boundingBox.bndbox[0].xmax[0],
    ymin: boundingBox.bndbox[0].ymin[0],
    ymax: boundingBox.bndbox[0].ymax[0],
  };
};

const makeDir = function(dirName: string) {
  fs.mkdirSync(dirName, { recursive: true }, err => {
    if (err) {
      throw new CliTerseError(err);
    }
  });
};

export const datasetResize = CliLeaf({
  name: 'resize',
  description: 'Resize images to the input size of model training',
  namedInputs: {
    'target-dir': CliStringInput({
      required: true,
      placeholder: '</path/to/target/directory>',
    }),
    'output-dir': CliStringInput({
      required: false,
      placeholder: '</path/to/output/directory>',
    }),
    width: CliNumberInput({
      required: false,
      placeholder: `Width to resize to. Default is set to ${PixelWidth}.`,
    }),
    height: CliNumberInput({
      required: false,
      placeholder: `Height to resize to. Default is set to ${PixelHeight}.`,
    }),
  },

  async action(_, opts) {
    const targetDir = opts['target-dir'];
    if (!targetDir.endsWith('.zip')) {
      throw new CliTerseError('Target directory needs to be a zip file!');
    }

    const extractedDir = 'extractedDir';
    const unzip = new admzip(targetDir);
    unzip.extractAllTo(extractedDir);

    const imageDir = `${extractedDir}/${ImageDirName}`;
    const annotDir = `${extractedDir}/${AnnotationDirName}`;

    const outputDir = opts['output-dir'] || `resized_${path.basename(targetDir)}`;
    if (!outputDir.endsWith('.zip')) {
      throw new CliTerseError('output-dir needs to be suffixed with .zip!');
    }

    const tmpDir = 'tmp';
    const outputImages = `${tmpDir}/${ImageDirName}`;
    const outputAnnots = `${tmpDir}/${AnnotationDirName}`;

    const width = opts['width'] || PixelWidth;
    const height = opts['height'] || PixelHeight;

    const builder = new xml2js.Builder({ headless: true });

    makeDir(outputImages);
    makeDir(outputAnnots);

    winston.info('Resizing images and annotations.');
    // Resize Images
    const sharpPromises: any = [];
    const imgEnts = await fs.readdirSync(imageDir, { withFileTypes: true });
    imgEnts
      .filter(entry => entry.isFile())
      .filter(entry => entry.name.endsWith('.jpg'))
      .map(entry => entry.name)
      .forEach(image => {
        const sharpPromise = sharp(`${imageDir}/${image}`)
          .resize(width, height, {
            kernal: 'nearest',
            fit: 'fill',
          })
          .toFile(`${outputImages}/${image}`);
        sharpPromises.push(sharpPromise);
      });
    await Promise.all(sharpPromises);

    // Resize Annotations
    const annotEnts = await fs.readdirSync(annotDir, { withFileTypes: true });
    annotEnts
      .filter(entry => entry.isFile())
      .filter(entry => entry.name.endsWith('.xml'))
      .map(entry => entry.name)
      .forEach(annot => {
        const data = fs.readFileSync(`${annotDir}/${annot}`);
        xml2js.parseString(data, (err, xml) => {
          if (err) {
            throw new CliTerseError(err);
          }

          const ogHeight = Number(xml.annotation.size[0].height[0]);
          const ogWidth = Number(xml.annotation.size[0].width[0]);

          const xScale = width / ogWidth;
          const yScale = height / ogHeight;

          xml.annotation.object.forEach(box => {
            const parsed = parseBoundingBox(box);

            box.bndbox[0].xmin[0] = parsed.xmin * xScale;
            box.bndbox[0].xmax[0] = parsed.xmax * xScale;
            box.bndbox[0].ymin[0] = parsed.ymin * yScale;
            box.bndbox[0].ymax[0] = parsed.ymax * yScale;
          });

          const resizedXml = builder.buildObject(xml);

          fs.writeFileSync(`${outputAnnots}/${annot}`, resizedXml);
        });
      });
    // Zip resized images and annotations
    const zip = new admzip();
    zip.addLocalFolder(tmpDir);
    zip.writeZip(outputDir);
    // Remove temp dir and extracted dir
    rimraf.sync(extractedDir);
    rimraf.sync(tmpDir);
    winston.info('Done!');
  },
});
