import { dirname, basename } from 'path';
import { stat, createReadStream } from 'fs';
import fetch, { Response } from 'node-fetch';
import { CLOUD_API_MODEL_VERSION_PACKAGES_PATH, CloudApiUrl } from '@alwaysai/cloud-api';
import { CodedError } from '@carnesen/coded-error';
import * as tar from 'tar';
const path = require('path');
const fs = require('fs');
import rimraf = require('rimraf');

import { CliRpcClient } from './rpc-client';
import { CliAuthenticationClient } from './authentication-client';
import { ModelJsonFile } from './model-json-file';
import { promisify } from 'util';
import tempy = require('tempy');
import { getSystemId } from './system-id';
import { ModelPackageDirectory } from './model-package-cache';

import * as winston from 'winston';
import { downloadModelPackageToCache } from './download-model-package-to-cache';

function compareModelVersions(a, b) {
  const aVersion = parseFloat(a.split('.tar')[0]);
  const bVersion = parseFloat(b.split('.tar')[0]);
  if (aVersion < bVersion) {
    return -1;
  }
  if (aVersion > bVersion) {
    return 1;
  }
  return 0;
}

function getLatestModel(id: string) {
  const modelDirectory = path.join(ModelPackageDirectory(), id);
  let modelPathList;
  try {
    modelPathList = fs.readdirSync(modelDirectory);
  } catch (error) {
    throw new Error(`Model ${id} not found in cache`);
  }
  modelPathList.sort(compareModelVersions);
  const latestModel = modelPathList.pop();
  return path.join(ModelPackageDirectory(), id, latestModel);
}

function getVersionFromModelPath(path_: string) {
  return parseFloat(path.basename(path_).split('.tar')[0]);
}

async function extractConfigFromCachedModel(id: string, destDir: string) {
  const lateModelPath = getLatestModel(id);
  await tar.extract({ file: lateModelPath, cwd: destDir });
}

export const modelPackageCloudClient = {
  async download(id: string, version: number) {
    const { uuid } = await CliRpcClient().getModelVersion({ id, version });
    const modelPackageUrl = ModelPackageUrl(uuid);
    const authorizationHeader = await CliAuthenticationClient().getAuthorizationHeader();
    const response = await fetch(modelPackageUrl, {
      method: 'GET',
      headers: { ...authorizationHeader, 'Content-Length': '0' },
    });
    await throwIfNotOk(response);
    return response.body;
  },

  async publish(id: string | undefined) {
    let modelPackagePath;
    let modelJson;
    let originalVersion;
    let id_ = id;
    if (!id) {
      // Get the model config from current dir
      // tar the package
      const modelConfigDir = process.cwd();
      modelJson = ModelJsonFile(modelConfigDir);
      modelPackagePath = tempy.file();
      await tar.create(
        { cwd: dirname(modelConfigDir), gzip: true, file: modelPackagePath },
        [basename(process.cwd())],
      );
      id_ = modelJson.id;
    } else {
      // Extract model config to temp dir and read
      // Define latest model for id as model package to publish
      const tempDir = tempy.directory();
      const modelConfigDir = path.join(tempDir, id.split('/').pop());
      await extractConfigFromCachedModel(id, tempDir);
      modelJson = ModelJsonFile(modelConfigDir).read();
      rimraf.sync(tempDir);
      modelPackagePath = getLatestModel(id);
      originalVersion = getVersionFromModelPath(modelPackagePath);
      id_ = id;
    }
    const rpcClient = CliRpcClient();
    const { uuid } = await rpcClient.createModelVersion(modelJson);
    const authorizationHeader = await CliAuthenticationClient().getAuthorizationHeader();
    const stats = await promisify(stat)(modelPackagePath);
    const modelPackageStream = createReadStream(modelPackagePath);
    const response = await fetch(ModelPackageUrl(uuid), {
      method: 'PUT',
      headers: {
        'Content-Type': 'application/gzip',
        'Content-Length': stats.size.toString(),
        ...authorizationHeader,
      },
      body: modelPackageStream,
    });

    await throwIfNotOk(response);
    await rpcClient.finalizeModelVersion(uuid);
    const { version: latestVersion } = await CliRpcClient().getModelVersion({ id: id_! });
    if (!id) {
      winston.info(`Published ${id_} version ${latestVersion}`);
    } else {
      winston.info(
        `Publishing from cache ${id_} version ${originalVersion}, Published to model catalog as ${id_} version ${latestVersion}`,
      );
    }
    await downloadModelPackageToCache(id_!, latestVersion);
    return uuid;
  },
};

function ModelPackageUrl(uuid: string) {
  const cloudApiUrl = CloudApiUrl(getSystemId());
  return `${cloudApiUrl}${CLOUD_API_MODEL_VERSION_PACKAGES_PATH}/${uuid}`;
}

async function throwIfNotOk(response: Response) {
  if (!response.ok) {
    const parsed = await response.json();
    const {
      message = `Server responded ${response.status} ("${response.statusText}")`,
      code,
      data,
    } = parsed;
    throw new CodedError(message, code, data);
  }
}
